import pytest

from vellum.workflows.descriptors.exceptions import InvalidExpressionException
from vellum.workflows.references.constant import ConstantValueReference
from vellum.workflows.state.base import BaseState


def test_parse_json_invalid_json():
    # GIVEN an invalid JSON string
    state = BaseState()
    expression = ConstantValueReference('{"key": value}').parse_json()

    # WHEN we attempt to resolve the expression
    with pytest.raises(InvalidExpressionException) as exc_info:
        expression.resolve(state)

    # THEN an exception should be raised
    assert "Failed to parse JSON" in str(exc_info.value)


def test_parse_json_invalid_type():
    # GIVEN a non-string value
    state = BaseState()
    expression = ConstantValueReference(123).parse_json()

    # WHEN we attempt to resolve the expression
    with pytest.raises(InvalidExpressionException) as exc_info:
        expression.resolve(state)

    # THEN an exception should be raised
    assert "Expected a string, but got 123 of type <class 'int'>" == str(exc_info.value)



import pytest
from uuid import uuid4
from typing import Type

from vellum.workflows.types.core import JsonObject
from vellum.workflows.types.generics import NodeType
from vellum_ee.workflows.display.base import WorkflowMetaDisplay
from vellum_ee.workflows.display.editor.types import NodeDisplayData
from vellum_ee.workflows.display.nodes.get_node_display_class import get_node_display_class
from vellum_ee.workflows.display.types import (
    NodeDisplays,
    NodeOutputDisplays,
    StateValueDisplays,
    WorkflowDisplayContext,
    WorkflowInputsDisplays,
)
from vellum_ee.workflows.display.workflows.base_workflow_display import BaseWorkflowDisplay


@pytest.fixture()
def serialize_node():
    def _serialize_node(
        node_class: Type[NodeType],
        global_workflow_input_displays: WorkflowInputsDisplays = {},
        global_state_value_displays: StateValueDisplays = {},
        global_node_displays: NodeDisplays = {},
        global_node_output_displays: NodeOutputDisplays = {},
    ) -> JsonObject:
        node_display_class = get_node_display_class(node_class)
        node_display = node_display_class()

        context: WorkflowDisplayContext = WorkflowDisplayContext(
            workflow_display_class=BaseWorkflowDisplay,
            workflow_display=WorkflowMetaDisplay(
                entrypoint_node_id=uuid4(),
                entrypoint_node_source_handle_id=uuid4(),
                entrypoint_node_display=NodeDisplayData(),
            ),
            node_displays={node_class: node_display},
            global_workflow_input_displays=global_workflow_input_displays,
            global_state_value_displays=global_state_value_displays,
            global_node_displays=global_node_displays,
            global_node_output_displays=global_node_output_displays,
        )
        return node_display.serialize(context)

    return _serialize_node



import uuid

from vellum.client.types.string_vellum_value import StringVellumValue
from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes import InlinePromptNode
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.nodes.mocks import MockNodeExecution
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.types import NodeOutputDisplay


def test_mocks__parse_from_app():
    # GIVEN a PromptNode
    class PromptNode(InlinePromptNode):
        pass

    # AND a workflow class with that PromptNode
    class MyWorkflow(BaseWorkflow):
        graph = PromptNode

    # AND a mock workflow node execution from the app
    raw_mock_workflow_node_execution = [
        {
            "type": "WORKFLOW_NODE_OUTPUT",
            "node_id": str(PromptNode.__id__),
            "mock_executions": [
                {
                    "when_condition": {
                        "expression": {
                            "type": "LOGICAL_CONDITION_GROUP",
                            "combinator": "AND",
                            "negated": False,
                            "conditions": [
                                {
                                    "type": "LOGICAL_CONDITION",
                                    "lhs_variable_id": "e60902d5-6892-4916-80c1-f0130af52322",
                                    "operator": ">=",
                                    "rhs_variable_id": "5c1bbb24-c288-49cb-a9b7-0c6f38a86037",
                                }
                            ],
                        },
                        "variables": [
                            {
                                "type": "EXECUTION_COUNTER",
                                "node_id": str(PromptNode.__id__),
                                "id": "e60902d5-6892-4916-80c1-f0130af52322",
                            },
                            {
                                "type": "CONSTANT_VALUE",
                                "variable_value": {"type": "NUMBER", "value": 0},
                                "id": "5c1bbb24-c288-49cb-a9b7-0c6f38a86037",
                            },
                        ],
                    },
                    "then_outputs": [
                        {
                            "output_id": "9e6dc5d3-8ea0-4346-8a2a-7cce5495755b",
                            "value": {
                                "id": "27006b2a-fa81-430c-a0b2-c66a9351fc68",
                                "type": "CONSTANT_VALUE",
                                "variable_value": {"type": "STRING", "value": "Hello"},
                            },
                        },
                        {
                            "output_id": "60305ffd-60b0-42aa-b54e-4fdae0f8c28a",
                            "value": {
                                "id": "4559c778-6e27-4cfe-a460-734ba62a5082",
                                "type": "CONSTANT_VALUE",
                                "variable_value": {"type": "ARRAY", "value": [{"type": "STRING", "value": "Hello"}]},
                            },
                        },
                    ],
                }
            ],
        }
    ]

    # WHEN we parse the mock workflow node execution
    node_output_mocks = MockNodeExecution.validate_all(
        raw_mock_workflow_node_execution,
        MyWorkflow,
    )

    # THEN we get a list of MockNodeExecution objects
    assert node_output_mocks
    assert len(node_output_mocks) == 1
    assert node_output_mocks[0] == MockNodeExecution(
        when_condition=PromptNode.Execution.count.greater_than_or_equal_to(0.0),
        then_outputs=PromptNode.Outputs(
            text="Hello",
            results=[
                StringVellumValue(value="Hello"),
            ],
        ),
    )


def test_mocks__parse_none_still_runs():
    # GIVEN a Base Node
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND a workflow class with that Node
    class MyWorkflow(BaseWorkflow):
        graph = StartNode

        class Outputs(BaseWorkflow.Outputs):
            final_value = StartNode.Outputs.foo

    # AND we parsed `None` on `MockNodeExecution`
    node_output_mocks = MockNodeExecution.validate_all(
        None,
        MyWorkflow,
    )

    # WHEN we run the workflow
    workflow = MyWorkflow()
    final_event = workflow.run(node_output_mocks=node_output_mocks)

    # THEN it was successful
    assert final_event.name == "workflow.execution.fulfilled"


def test_mocks__use_id_from_display():
    # GIVEN a Base Node
    class StartNode(BaseNode):
        class Outputs(BaseNode.Outputs):
            foo: str

    # AND a workflow class with that Node
    class MyWorkflow(BaseWorkflow):
        graph = StartNode

        class Outputs(BaseWorkflow.Outputs):
            final_value = StartNode.Outputs.foo

    # AND a display class on that Base Node
    node_output_id = uuid.uuid4()

    class StartNodeDisplay(BaseNodeDisplay[StartNode]):
        output_display = {StartNode.Outputs.foo: NodeOutputDisplay(id=node_output_id, name="foo")}

    # AND a mock workflow node execution from the app
    raw_mock_workflow_node_execution = [
        {
            "type": "WORKFLOW_NODE_OUTPUT",
            "node_id": str(StartNode.__id__),
            "mock_executions": [
                {
                    "when_condition": {
                        "expression": {
                            "type": "LOGICAL_CONDITION_GROUP",
                            "combinator": "AND",
                            "negated": False,
                            "conditions": [
                                {
                                    "type": "LOGICAL_CONDITION",
                                    "lhs_variable_id": "e60902d5-6892-4916-80c1-f0130af52322",
                                    "operator": ">=",
                                    "rhs_variable_id": "5c1bbb24-c288-49cb-a9b7-0c6f38a86037",
                                }
                            ],
                        },
                        "variables": [
                            {
                                "type": "EXECUTION_COUNTER",
                                "node_id": str(StartNode.__id__),
                                "id": "e60902d5-6892-4916-80c1-f0130af52322",
                            },
                            {
                                "type": "CONSTANT_VALUE",
                                "variable_value": {"type": "NUMBER", "value": 0},
                                "id": "5c1bbb24-c288-49cb-a9b7-0c6f38a86037",
                            },
                        ],
                    },
                    "then_outputs": [
                        {
                            "output_id": str(node_output_id),
                            "value": {
                                "id": "27006b2a-fa81-430c-a0b2-c66a9351fc68",
                                "type": "CONSTANT_VALUE",
                                "variable_value": {"type": "STRING", "value": "Hello"},
                            },
                        },
                    ],
                }
            ],
        }
    ]

    # WHEN we parsed the raw data on `MockNodeExecution`
    node_output_mocks = MockNodeExecution.validate_all(
        raw_mock_workflow_node_execution,
        MyWorkflow,
    )

    # THEN we get the expected list of MockNodeExecution objects
    assert node_output_mocks
    assert len(node_output_mocks) == 1
    assert node_output_mocks[0] == MockNodeExecution(
        when_condition=StartNode.Execution.count.greater_than_or_equal_to(0.0),
        then_outputs=StartNode.Outputs(
            foo="Hello",
        ),
    )



# WARNING: This file will be removed in a future release. Please import from "vellum.client" instead.

from vellum.client.types.test_suite_run_prompt_sandbox_history_item_exec_config_data_request import *



# This file was auto-generated by Fern from our API Definition.

import typing

from .reducto_chunking_request import ReductoChunkingRequest
from .sentence_chunking_request import SentenceChunkingRequest
from .token_overlapping_window_chunking_request import TokenOverlappingWindowChunkingRequest

DocumentIndexChunkingRequest = typing.Union[
    ReductoChunkingRequest, SentenceChunkingRequest, TokenOverlappingWindowChunkingRequest
]



from uuid import UUID
from typing import Any, ClassVar, Generic, Optional, TypeVar

from vellum.workflows.nodes import ErrorNode
from vellum.workflows.types.core import JsonArray, JsonObject
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.utils import raise_if_descriptor
from vellum_ee.workflows.display.nodes.vellum.utils import create_node_input
from vellum_ee.workflows.display.types import WorkflowDisplayContext

_ErrorNodeType = TypeVar("_ErrorNodeType", bound=ErrorNode)
LEGACY_INPUT_NAME = "error_source_input_id"


class BaseErrorNodeDisplay(BaseNodeDisplay[_ErrorNodeType], Generic[_ErrorNodeType]):
    # DEPRECATED: Remove in 0.15.0 once removed from the vellum-side
    error_output_id: ClassVar[Optional[UUID]] = None
    # DEPRECATED: Remove in 0.15.0 once removed from the vellum-side
    name: ClassVar[Optional[str]] = None

    __serializable_inputs__ = {ErrorNode.error}

    def serialize_ports(self, display_context: "WorkflowDisplayContext") -> JsonArray:
        """Error nodes have no ports."""
        return []

    def serialize(self, display_context: WorkflowDisplayContext, **_kwargs) -> JsonObject:
        node_id = self.node_id
        error_source_input_id = self.node_input_ids_by_name.get(
            ErrorNode.error.name,
        ) or self.node_input_ids_by_name.get(LEGACY_INPUT_NAME)

        error_attribute = raise_if_descriptor(self._node.error)

        node_inputs = [
            create_node_input(
                node_id=node_id,
                input_name=LEGACY_INPUT_NAME,
                value=error_attribute,
                display_context=display_context,
                input_id=error_source_input_id,
            )
        ]

        node_data: dict[str, Any] = {
            "id": str(node_id),
            "type": "ERROR",
            "inputs": [node_input.dict() for node_input in node_inputs],
            "data": {
                "label": self.label,
                "target_handle_id": str(self.get_target_handle_id()),
                "error_source_input_id": str(error_source_input_id),
            },
            **self.serialize_generic_fields(display_context),
        }

        if self.name:
            node_data["data"]["name"] = self.name

        if self.error_output_id:
            node_data["data"]["error_output_id"] = str(self.error_output_id)

        return node_data



import subprocess
from unittest.mock import patch
from uuid import uuid4

from click.testing import CliRunner

from vellum.client.core.api_error import ApiError
from vellum_cli import main as cli_main


@patch("subprocess.run")
@patch("docker.from_env")
def test_image_push_docker_service_token_401_error(mock_docker_from_env, mock_run, vellum_client, monkeypatch):
    monkeypatch.setenv("VELLUM_API_URL", "https://api.vellum.ai")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    mock_docker_client = mock_docker_from_env.return_value
    mock_docker_client.images.get.return_value.id = "test-image-id"

    mock_run.side_effect = [
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Pruning successful"),
        subprocess.CompletedProcess(
            args="", returncode=0, stdout=b'{"manifests": [{"platform": {"architecture": "amd64"}}]}'
        ),
    ]

    vellum_client.container_images.docker_service_token.side_effect = ApiError(status_code=401, body="Unauthorized")

    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myimage:latest"])

    assert result.exit_code == 1
    assert "Authentication failed" in result.output
    assert "VELLUM_API_KEY" in result.output


@patch("subprocess.run")
@patch("docker.from_env")
def test_image_push_docker_service_token_500_error(mock_docker_from_env, mock_run, vellum_client, monkeypatch):
    monkeypatch.setenv("VELLUM_API_URL", "https://api.vellum.ai")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    mock_docker_client = mock_docker_from_env.return_value
    mock_docker_client.images.get.return_value.id = "test-image-id"

    mock_run.side_effect = [
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Pruning successful"),
        subprocess.CompletedProcess(
            args="", returncode=0, stdout=b'{"manifests": [{"platform": {"architecture": "amd64"}}]}'
        ),
    ]

    vellum_client.container_images.docker_service_token.side_effect = ApiError(
        status_code=500, body="Internal Server Error"
    )

    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myimage:latest"])

    assert result.exit_code == 1
    assert "Server error" in result.output
    assert "try again later" in result.output


@patch("subprocess.run")
@patch("docker.from_env")
def test_image_push_docker_service_token_other_error(mock_docker_from_env, mock_run, vellum_client, monkeypatch):
    monkeypatch.setenv("VELLUM_API_URL", "https://api.vellum.ai")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    mock_docker_client = mock_docker_from_env.return_value
    mock_docker_client.images.get.return_value.id = "test-image-id"

    mock_run.side_effect = [
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Pruning successful"),
        subprocess.CompletedProcess(
            args="", returncode=0, stdout=b'{"manifests": [{"platform": {"architecture": "amd64"}}]}'
        ),
    ]

    vellum_client.container_images.docker_service_token.side_effect = ApiError(
        status_code=429, body="Too Many Requests"
    )

    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myimage:latest"])

    assert result.exit_code == 1
    assert "API request failed" in result.output
    assert "HTTP 429" in result.output


@patch("subprocess.run")
@patch("docker.from_env")
def test_image_push_container_image_401_error(mock_docker_from_env, mock_run, vellum_client, monkeypatch):
    monkeypatch.setenv("VELLUM_API_URL", "https://api.vellum.ai")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    mock_docker_client = mock_docker_from_env.return_value
    mock_docker_client.images.get.return_value.id = "test-image-id"
    mock_docker_client.images.push.return_value = ["pushed"]

    mock_run.side_effect = [
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Pruning successful"),
        subprocess.CompletedProcess(
            args="", returncode=0, stdout=b'{"manifests": [{"platform": {"architecture": "amd64"}}]}'
        ),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b'[{"RepoDigests": ["test-repo@sha256:abcd1234"]}]'),
    ]

    vellum_client.container_images.docker_service_token.return_value = type(
        "obj", (object,), {"access_token": "345678mnopqr", "organization_id": str(uuid4()), "repository": "myrepo.net"}
    )()

    vellum_client.container_images.push_container_image.side_effect = ApiError(status_code=401, body="Unauthorized")

    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myimage:latest"])

    assert result.exit_code == 1
    assert "Authentication failed" in result.output
    assert "VELLUM_API_KEY" in result.output


@patch("subprocess.run")
@patch("docker.from_env")
def test_image_push_container_image_500_error(mock_docker_from_env, mock_run, vellum_client, monkeypatch):
    monkeypatch.setenv("VELLUM_API_URL", "https://api.vellum.ai")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    mock_docker_client = mock_docker_from_env.return_value
    mock_docker_client.images.get.return_value.id = "test-image-id"
    mock_docker_client.images.push.return_value = ["pushed"]

    mock_run.side_effect = [
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Pruning successful"),
        subprocess.CompletedProcess(
            args="", returncode=0, stdout=b'{"manifests": [{"platform": {"architecture": "amd64"}}]}'
        ),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b'[{"RepoDigests": ["test-repo@sha256:abcd1234"]}]'),
    ]

    vellum_client.container_images.docker_service_token.return_value = type(
        "obj", (object,), {"access_token": "345678mnopqr", "organization_id": str(uuid4()), "repository": "myrepo.net"}
    )()

    vellum_client.container_images.push_container_image.side_effect = ApiError(
        status_code=500, body="Internal Server Error"
    )

    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myimage:latest"])

    assert result.exit_code == 1
    assert "Server error" in result.output
    assert "try again later" in result.output


@patch("subprocess.run")
@patch("docker.from_env")
def test_image_push_container_image_other_error(mock_docker_from_env, mock_run, vellum_client, monkeypatch):
    monkeypatch.setenv("VELLUM_API_URL", "https://api.vellum.ai")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    mock_docker_client = mock_docker_from_env.return_value
    mock_docker_client.images.get.return_value.id = "test-image-id"
    mock_docker_client.images.push.return_value = ["pushed"]

    mock_run.side_effect = [
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Pruning successful"),
        subprocess.CompletedProcess(
            args="", returncode=0, stdout=b'{"manifests": [{"platform": {"architecture": "amd64"}}]}'
        ),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b'[{"RepoDigests": ["test-repo@sha256:abcd1234"]}]'),
    ]

    vellum_client.container_images.docker_service_token.return_value = type(
        "obj", (object,), {"access_token": "345678mnopqr", "organization_id": str(uuid4()), "repository": "myrepo.net"}
    )()

    vellum_client.container_images.push_container_image.side_effect = ApiError(
        status_code=429, body="Too Many Requests"
    )

    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myimage:latest"])

    assert result.exit_code == 1
    assert "API request failed" in result.output
    assert "HTTP 429" in result.output



import importlib
import inspect
from types import FrameType
from uuid import UUID
from typing import Annotated, Any, Dict, Literal, Optional, Union

from pydantic import BeforeValidator

from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.client.types.code_resource_definition import CodeResourceDefinition as ClientCodeResourceDefinition
from vellum.workflows.constants import AuthorizationType
from vellum.workflows.references.environment_variable import EnvironmentVariableReference


def serialize_type_encoder(obj: type) -> Dict[str, Any]:
    return {
        "name": obj.__name__,
        "module": obj.__module__.split("."),
    }


def serialize_type_encoder_with_id(obj: Union[type, "CodeResourceDefinition"]) -> Dict[str, Any]:
    if hasattr(obj, "__id__") and isinstance(obj, type):
        return {
            "id": getattr(obj, "__id__"),
            **serialize_type_encoder(obj),
        }
    elif isinstance(obj, CodeResourceDefinition):
        return obj.model_dump(mode="json")

    raise AttributeError(f"The object of type '{type(obj).__name__}' must have an '__id__' attribute.")


class CodeResourceDefinition(ClientCodeResourceDefinition):
    id: UUID

    @staticmethod
    def encode(obj: type) -> "CodeResourceDefinition":
        return CodeResourceDefinition(**serialize_type_encoder_with_id(obj))

    def decode(self) -> Any:
        if ".<locals>." in self.name:
            # We are decoding a local class that should already be loaded in our stack frame. So
            # we climb up to look for it.
            frame = inspect.currentframe()
            return self._resolve_local(frame)

        try:
            imported_module = importlib.import_module(".".join(self.module))
        except ImportError:
            return None

        return getattr(imported_module, self.name, None)

    def _resolve_local(self, frame: Optional[FrameType]) -> Any:
        if not frame:
            return None

        frame_module = frame.f_globals.get("__name__")
        if not isinstance(frame_module, str) or frame_module.split(".") != self.module:
            return self._resolve_local(frame.f_back)

        outer, inner = self.name.split(".<locals>.")
        frame_outer = frame.f_code.co_name
        if frame_outer != outer:
            return self._resolve_local(frame.f_back)

        return frame.f_locals.get(inner)


VellumCodeResourceDefinition = Annotated[
    CodeResourceDefinition,
    BeforeValidator(lambda d: (d if type(d) is dict else serialize_type_encoder_with_id(d))),
]


class DeploymentDefinition(UniversalBaseModel):
    deployment: str
    release_tag: str = "LATEST"

    def _is_uuid(self) -> bool:
        """Check if the deployment field is a valid UUID."""
        try:
            UUID(self.deployment)
            return True
        except ValueError:
            return False

    @property
    def deployment_id(self) -> Optional[UUID]:
        """Get the deployment ID if the deployment field is a UUID."""
        if self._is_uuid():
            return UUID(self.deployment)
        return None

    @property
    def deployment_name(self) -> Optional[str]:
        """Get the deployment name if the deployment field is not a UUID."""
        if not self._is_uuid():
            return self.deployment
        return None


class ComposioToolDefinition(UniversalBaseModel):
    """Represents a specific Composio action that can be used in Tool Calling Node"""

    type: Literal["COMPOSIO"] = "COMPOSIO"

    # Core identification
    toolkit: str  # "GITHUB", "SLACK", etc.
    action: str  # Specific action like "GITHUB_CREATE_AN_ISSUE"
    description: str
    user_id: Optional[str] = None
    name: str = ""

    def model_post_init(self, __context: Any):
        if self.name == "":
            self.name = self.action.lower()


class MCPServer(UniversalBaseModel):
    type: Literal["MCP_SERVER"] = "MCP_SERVER"
    name: str
    url: str
    authorization_type: Optional[AuthorizationType] = None
    bearer_token_value: Optional[Union[str, EnvironmentVariableReference]] = None
    api_key_header_key: Optional[str] = None
    api_key_header_value: Optional[Union[str, EnvironmentVariableReference]] = None

    model_config = {"arbitrary_types_allowed": True}

    def __setattr__(self, name: str, value: Any) -> None:
        """Override to automatically set serialization flags for environment variables."""
        super().__setattr__(name, value)

        if name == "bearer_token_value" and isinstance(value, EnvironmentVariableReference):
            value.serialize_as_constant = True

        if name == "api_key_header_value" and isinstance(value, EnvironmentVariableReference):
            value.serialize_as_constant = True


class MCPToolDefinition(UniversalBaseModel):
    name: str
    server: MCPServer
    description: Optional[str] = None
    parameters: Dict[str, Any] = {}



from .environment_variable import EnvironmentVariableReference
from .external_input import ExternalInputReference
from .lazy import LazyReference
from .node import NodeReference
from .output import OutputReference
from .state_value import StateValueReference
from .vellum_secret import VellumSecretReference
from .workflow_input import WorkflowInputReference

__all__ = [
    "EnvironmentVariableReference",
    "ExternalInputReference",
    "LazyReference",
    "NodeReference",
    "OutputReference",
    "StateValueReference",
    "VellumSecretReference",
    "WorkflowInputReference",
]



import json
import logging
import os
from uuid import uuid4
from typing import Any, Iterable, Iterator, List, Literal, Union, cast

from openai import OpenAI
from openai.types.chat import (
    ChatCompletionAssistantMessageParam,
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartInputAudioParam,
    ChatCompletionContentPartParam,
    ChatCompletionContentPartRefusalParam,
    ChatCompletionContentPartTextParam,
    ChatCompletionMessageParam,
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
)
from openai.types.chat.chat_completion_chunk import Choice

from vellum import (
    AdHocExecutePromptEvent,
    FulfilledAdHocExecutePromptEvent,
    InitiatedAdHocExecutePromptEvent,
    RejectedAdHocExecutePromptEvent,
    StreamingAdHocExecutePromptEvent,
    StringVellumValue,
    VellumAudio,
    VellumError,
    VellumImage,
)
from vellum.prompts.blocks.compilation import compile_prompt_blocks
from vellum.prompts.blocks.types import CompiledChatMessagePromptBlock
from vellum.workflows.errors import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes import InlinePromptNode
from vellum.workflows.types.generics import StateType

logger = logging.getLogger(__name__)


class OpenAIChatCompletionNode(InlinePromptNode[StateType]):
    """
    Used to execute a Prompt using the OpenAI API.
    """

    # Override
    def _get_prompt_event_stream(self) -> Iterator[AdHocExecutePromptEvent]:
        client = self._get_client()

        execution_id = str(uuid4())

        yield InitiatedAdHocExecutePromptEvent(
            execution_id=execution_id,
        )

        try:
            stream = client.chat.completions.create(
                messages=self._get_messages(),
                model=self.ml_model,
                # TODO: Add support for additional parameters
                stream=True,
            )
        except Exception as exc:
            yield RejectedAdHocExecutePromptEvent(
                error=VellumError(
                    code=WorkflowErrorCode.PROVIDER_ERROR,
                    message=exc.args[0],
                ),
                execution_id=execution_id,
            )
            return

        combined_delta_content = ""
        for chunk in stream:
            choices: List[Choice] = chunk.choices
            if len(choices) != 1:
                yield RejectedAdHocExecutePromptEvent(
                    error=VellumError(
                        code=WorkflowErrorCode.PROVIDER_ERROR,
                        message="Expected one choice per chunk, but found more than one.",
                    ),
                    execution_id=execution_id,
                )
                return

            choice = choices[0]
            delta = choice.delta

            if delta.tool_calls:
                # TODO: Add support for tool calls
                raise NotImplementedError("This node hasn't been extended to support tool calling yet.")

            if delta.content:
                combined_delta_content += delta.content

            StreamingAdHocExecutePromptEvent(
                output=StringVellumValue(value=delta.content),
                # TODO: Add support for multiple outputs
                output_index=1,
                execution_id=execution_id,
            )

        yield FulfilledAdHocExecutePromptEvent(
            # TODO: Add support for multiple outputs
            outputs=[
                StringVellumValue(value=combined_delta_content),
            ],
            execution_id=execution_id,
        )

    def _get_client(self) -> OpenAI:
        """Used to retrieve an API client for interacting with the OpenAI API.

        Note: This method can be overridden if you'd like to use your own API client that conforms to the same
            interfaces as that of OpenAI.
        """

        openai_api_key = os.environ.get("OPENAI_API_KEY")

        if not openai_api_key:
            raise NodeException(
                code=WorkflowErrorCode.INTERNAL_ERROR,
                message="Unable to determine an OpenAI API key.",
            )

        client = OpenAI(api_key=openai_api_key)
        return client

    def _get_messages(self) -> Iterable[ChatCompletionMessageParam]:
        input_variables, input_values = self._compile_prompt_inputs()

        compiled_blocks = compile_prompt_blocks(
            blocks=self.blocks, inputs=input_values, input_variables=input_variables
        )

        chat_message_blocks: list[CompiledChatMessagePromptBlock] = [
            block for block in compiled_blocks if block.block_type == "CHAT_MESSAGE"
        ]
        messages = [self._create_message(block) for block in chat_message_blocks]

        return messages

    @classmethod
    def _create_message(cls, chat_message_block: CompiledChatMessagePromptBlock) -> ChatCompletionMessageParam:
        name = chat_message_block.source
        content = cls._create_message_content(chat_message_block)

        if chat_message_block.role == "SYSTEM":
            relevant_system_content = [
                cast(ChatCompletionContentPartTextParam, c) for c in content if c["type"] == "text"
            ]
            system_message: ChatCompletionSystemMessageParam = {
                "content": relevant_system_content,
                "role": "system",
            }
            if name:
                system_message["name"] = name

            return system_message
        elif chat_message_block.role == "USER":
            user_message: ChatCompletionUserMessageParam = {
                "content": content,
                "role": "user",
            }
            if name:
                user_message["name"] = name

            return user_message
        elif chat_message_block.role == "ASSISTANT":
            relevant_assistant_content = [
                cast(Union[ChatCompletionContentPartTextParam, ChatCompletionContentPartRefusalParam], c)
                for c in content
                if c["type"] in ["text", "refusal"]
            ]
            assistant_message: ChatCompletionAssistantMessageParam = {
                "content": relevant_assistant_content,
                "role": "assistant",
            }
            if name:
                assistant_message["name"] = name

            return assistant_message
        else:
            logger.error(f"Unexpected role: {chat_message_block.role}")
            raise NodeException(
                code=WorkflowErrorCode.INTERNAL_ERROR, message="Unexpected role found when compiling prompt blocks"
            )

    @classmethod
    def _create_message_content(
        cls,
        chat_message_block: CompiledChatMessagePromptBlock,
    ) -> List[ChatCompletionContentPartParam]:
        content: List[ChatCompletionContentPartParam] = []
        for block in chat_message_block.blocks:
            if block.content.type == "STRING":
                string_value = cast(str, block.content.value)
                string_content_item: ChatCompletionContentPartTextParam = {"type": "text", "text": string_value}
                content.append(string_content_item)
            elif block.content.type == "JSON":
                json_value = cast(Any, block.content.value)
                json_content_item: ChatCompletionContentPartTextParam = {"type": "text", "text": json.dumps(json_value)}
                content.append(json_content_item)
            elif block.content.type == "IMAGE":
                image_value = cast(VellumImage, block.content.value)
                image_content_item: ChatCompletionContentPartImageParam = {
                    "type": "image_url",
                    "image_url": {"url": image_value.src},
                }
                if image_value.metadata and image_value.metadata.get("detail"):
                    detail = image_value.metadata["detail"]

                    if detail not in ["auto", "low", "high"]:
                        raise NodeException(
                            code=WorkflowErrorCode.INTERNAL_ERROR,
                            message="Image detail must be one of 'auto', 'low', or 'high.",
                        )

                    image_content_item["image_url"]["detail"] = cast(Literal["auto", "low", "high"], detail)

                content.append(image_content_item)
            elif block.content.type == "AUDIO":
                audio_value = cast(VellumAudio, block.content.value)
                audio_value_src_parts = audio_value.src.split(",")
                if len(audio_value_src_parts) != 2:
                    raise NodeException(
                        code=WorkflowErrorCode.INTERNAL_ERROR, message="Audio data is not properly encoded."
                    )
                _, cleaned_audio_value = audio_value_src_parts
                if not audio_value.metadata:
                    raise NodeException(
                        code=WorkflowErrorCode.INTERNAL_ERROR, message="Audio metadata is required for audio input."
                    )
                audio_format = audio_value.metadata.get("format")
                if not audio_format:
                    raise NodeException(
                        code=WorkflowErrorCode.INTERNAL_ERROR, message="Audio format is required for audio input."
                    )
                if audio_format not in {"wav", "mp3"}:
                    raise NodeException(
                        code=WorkflowErrorCode.INTERNAL_ERROR,
                        message="Audio format must be one of 'wav' or 'mp3'.",
                    )

                audio_content_item: ChatCompletionContentPartInputAudioParam = {
                    "type": "input_audio",
                    "input_audio": {
                        "data": cleaned_audio_value,
                        "format": cast(Literal["wav", "mp3"], audio_format),
                    },
                }

                content.append(audio_content_item)
            elif block.content.type == "DOCUMENT":
                raise NodeException(
                    code=WorkflowErrorCode.PROVIDER_ERROR,
                    message="Document chat message content type is not currently supported",
                )
            else:
                raise NodeException(
                    code=WorkflowErrorCode.INTERNAL_ERROR,
                    message=f"Failed to parse chat message block {block.content.type}",
                )

        return content



import pytest

from deepdiff import DeepDiff

from vellum.workflows.expressions.begins_with import BeginsWithExpression
from vellum.workflows.expressions.between import BetweenExpression
from vellum.workflows.expressions.contains import ContainsExpression
from vellum.workflows.expressions.does_not_begin_with import DoesNotBeginWithExpression
from vellum.workflows.expressions.does_not_contain import DoesNotContainExpression
from vellum.workflows.expressions.does_not_end_with import DoesNotEndWithExpression
from vellum.workflows.expressions.does_not_equal import DoesNotEqualExpression
from vellum.workflows.expressions.ends_with import EndsWithExpression
from vellum.workflows.expressions.equals import EqualsExpression
from vellum.workflows.expressions.greater_than import GreaterThanExpression
from vellum.workflows.expressions.greater_than_or_equal_to import GreaterThanOrEqualToExpression
from vellum.workflows.expressions.in_ import InExpression
from vellum.workflows.expressions.is_not_null import IsNotNullExpression
from vellum.workflows.expressions.is_null import IsNullExpression
from vellum.workflows.expressions.less_than import LessThanExpression
from vellum.workflows.expressions.less_than_or_equal_to import LessThanOrEqualToExpression
from vellum.workflows.expressions.not_between import NotBetweenExpression
from vellum.workflows.expressions.not_in import NotInExpression
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_conditional_node.workflow import CategoryWorkflow
from tests.workflows.basic_conditional_node.workflow_with_only_one_conditional_node import create_simple_workflow


def test_serialize_workflow():
    # GIVEN a Workflow that uses a ConditionalNode
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=CategoryWorkflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1
    assert not DeepDiff(
        [
            {
                "id": "eece050a-432e-4a2c-8c87-9480397e4cbf",
                "key": "category",
                "type": "STRING",
                "required": True,
                "default": None,
                "extensions": {"color": None},
            },
        ],
        input_variables,
        ignore_order=True,
    )

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 5
    assert not DeepDiff(
        [
            {
                "id": "c05f7d96-59a0-4d58-93d7-d451afd3f630",
                "key": "question",
                "type": "STRING",
            },
            {
                "id": "93f2cb75-6fa2-4e46-9488-c0bcd29153c0",
                "key": "compliment",
                "type": "STRING",
            },
            {
                "id": "f936ae31-ba15-4864-8961-86231022a4d7",
                "key": "complaint",
                "type": "STRING",
            },
            {
                "id": "cdbe2adf-9951-409a-b9a8-b8b349037f4f",
                "key": "statement",
                "type": "STRING",
            },
            {
                "id": "62ad462f-f819-4940-99ab-b3f145507f57",
                "key": "fallthrough",
                "type": "STRING",
            },
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    assert len(workflow_raw_data["edges"]) == 11
    assert len(workflow_raw_data["nodes"]) == 12

    # AND each node should be serialized correctly
    entrypoint_node = workflow_raw_data["nodes"][0]
    assert entrypoint_node == {
        "id": "089b3201-537a-4ed7-8d15-2524a00e8534",
        "type": "ENTRYPOINT",
        "base": None,
        "definition": None,
        "inputs": [],
        "data": {
            "label": "Entrypoint Node",
            "source_handle_id": "c2f0871d-0d9d-417f-8b0e-c813ccf880ac",
        },
        "display_data": {
            "position": {"x": 0.0, "y": -50.0},
        },
    }

    conditional_node = workflow_raw_data["nodes"][1]
    assert not DeepDiff(
        {
            "id": "9b619e4d-b0a7-4121-9060-100d457868cb",
            "type": "CONDITIONAL",
            "inputs": [
                {
                    "id": "c99fb71e-a8ad-4627-b9a2-aa36bf1fdc02",
                    "key": "e0490a35-c863-422b-a12a-f18858d75241.field",
                    "value": {
                        "rules": [
                            {
                                "type": "INPUT_VARIABLE",
                                "data": {"input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "7aaebb17-b670-4366-80f9-fb569c8f8f85",
                    "key": "e0490a35-c863-422b-a12a-f18858d75241.value",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "question"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "d8675161-1aa1-4f10-8b98-ea6384211c87",
                    "key": "f7818d41-3f66-42c7-8218-de5c30379906.field",
                    "value": {
                        "rules": [
                            {
                                "type": "INPUT_VARIABLE",
                                "data": {"input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "4c77a541-9a13-41b2-b68e-ed1487d195d3",
                    "key": "f7818d41-3f66-42c7-8218-de5c30379906.value",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "complaint"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "a593293c-9c17-45c5-8c99-fc867aad580e",
                    "key": "38483416-c474-4e50-bfbc-7e92de2b4c2c.field",
                    "value": {
                        "rules": [
                            {
                                "type": "INPUT_VARIABLE",
                                "data": {"input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "d9ea5424-44d5-481a-b769-3bce5ef6c353",
                    "key": "38483416-c474-4e50-bfbc-7e92de2b4c2c.value",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "compliment"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "2066a9e9-7a55-4ef0-8897-7df32e31f0e8",
                    "key": "aa692069-9672-4adb-b874-74da98745f24.field",
                    "value": {
                        "rules": [
                            {
                                "type": "INPUT_VARIABLE",
                                "data": {"input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "fbbef65b-0d3b-4a10-9479-ebe07b8e3a82",
                    "key": "aa692069-9672-4adb-b874-74da98745f24.value",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "statement"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "3c1186a8-5a0f-4b6b-8637-9c09aea9c14d",
                    "key": "25aaa0ca-9312-49d7-9ad5-b6f40f8a0658.field",
                    "value": {
                        "rules": [
                            {
                                "type": "INPUT_VARIABLE",
                                "data": {"input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "78e99399-ed5a-4d26-8a42-c4f355cf60e3",
                    "key": "25aaa0ca-9312-49d7-9ad5-b6f40f8a0658.value",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "statement"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "5b7e4b08-68eb-471e-9eda-36c768f8eb46",
                    "key": "e59504e1-4134-46dc-8055-114c6a606af8.field",
                    "value": {
                        "rules": [
                            {
                                "type": "INPUT_VARIABLE",
                                "data": {"input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "02d26ab6-8faa-4da3-84b1-190581a6fd66",
                    "key": "e59504e1-4134-46dc-8055-114c6a606af8.value",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "statement"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
            ],
            "data": {
                "label": "Category Conditional Node",
                "target_handle_id": "dd89e228-a23e-422b-80b2-34362c1c050e",
                "conditions": [
                    {
                        "id": "de7b0b4e-7803-4d36-a275-2e7e3f60342b",
                        "type": "IF",
                        "source_handle_id": "3a45b81f-95e4-4cbd-8997-bfdbe30251e8",
                        "data": {
                            "id": "2ccd0730-26d1-4fb4-baa9-1a2a182dd9a0",
                            "rules": [
                                {
                                    "id": "e0490a35-c863-422b-a12a-f18858d75241",
                                    "rules": None,
                                    "combinator": None,
                                    "negated": False,
                                    "field_node_input_id": "c99fb71e-a8ad-4627-b9a2-aa36bf1fdc02",
                                    "operator": "=",
                                    "value_node_input_id": "7aaebb17-b670-4366-80f9-fb569c8f8f85",
                                }
                            ],
                            "combinator": "AND",
                            "negated": False,
                            "field_node_input_id": None,
                            "operator": None,
                            "value_node_input_id": None,
                        },
                    },
                    {
                        "id": "5e783d17-6808-441a-ac6c-33a4e184f4e0",
                        "type": "ELIF",
                        "source_handle_id": "7202f702-1ebc-4067-ab1e-ec67e49158ee",
                        "data": {
                            "id": "cc3f0d92-b603-42cc-b2e9-83e3b23b3bcb",
                            "rules": [
                                {
                                    "id": "f7818d41-3f66-42c7-8218-de5c30379906",
                                    "rules": None,
                                    "combinator": None,
                                    "negated": False,
                                    "field_node_input_id": "d8675161-1aa1-4f10-8b98-ea6384211c87",
                                    "operator": "=",
                                    "value_node_input_id": "4c77a541-9a13-41b2-b68e-ed1487d195d3",
                                }
                            ],
                            "combinator": "AND",
                            "negated": False,
                            "field_node_input_id": None,
                            "operator": None,
                            "value_node_input_id": None,
                        },
                    },
                    {
                        "id": "6bd2f643-9cf5-4e7f-9113-f90e5c8057be",
                        "type": "ELIF",
                        "source_handle_id": "cf45705d-1a47-43a6-9d24-a7fdf78baae0",
                        "data": {
                            "id": "a5a0f391-7052-452f-9fe1-a5781a491591",
                            "rules": [
                                {
                                    "id": "38483416-c474-4e50-bfbc-7e92de2b4c2c",
                                    "rules": None,
                                    "combinator": None,
                                    "negated": False,
                                    "field_node_input_id": "a593293c-9c17-45c5-8c99-fc867aad580e",
                                    "operator": "=",
                                    "value_node_input_id": "d9ea5424-44d5-481a-b769-3bce5ef6c353",
                                }
                            ],
                            "combinator": "AND",
                            "negated": False,
                            "field_node_input_id": None,
                            "operator": None,
                            "value_node_input_id": None,
                        },
                    },
                    {
                        "id": "0a058485-18a4-4e20-8a30-6da8196ac46f",
                        "type": "ELIF",
                        "source_handle_id": "f04610dd-61cf-41b0-b337-2235e101cdb0",
                        "data": {
                            "id": "efe7a851-2a67-4189-99ec-bc193242b270",
                            "rules": [
                                {
                                    "id": "aa692069-9672-4adb-b874-74da98745f24",
                                    "rules": None,
                                    "combinator": None,
                                    "negated": False,
                                    "field_node_input_id": "2066a9e9-7a55-4ef0-8897-7df32e31f0e8",
                                    "operator": "=",
                                    "value_node_input_id": "fbbef65b-0d3b-4a10-9479-ebe07b8e3a82",
                                },
                                {
                                    "id": "2c78817b-8b73-43fd-8dab-a8923018da9d",
                                    "rules": [
                                        {
                                            "id": "25aaa0ca-9312-49d7-9ad5-b6f40f8a0658",
                                            "rules": None,
                                            "combinator": None,
                                            "negated": False,
                                            "field_node_input_id": "3c1186a8-5a0f-4b6b-8637-9c09aea9c14d",
                                            "operator": "=",
                                            "value_node_input_id": "78e99399-ed5a-4d26-8a42-c4f355cf60e3",
                                        },
                                        {
                                            "id": "e59504e1-4134-46dc-8055-114c6a606af8",
                                            "rules": None,
                                            "combinator": None,
                                            "negated": False,
                                            "field_node_input_id": "5b7e4b08-68eb-471e-9eda-36c768f8eb46",
                                            "operator": "=",
                                            "value_node_input_id": "02d26ab6-8faa-4da3-84b1-190581a6fd66",
                                        },
                                    ],
                                    "combinator": "AND",
                                    "negated": False,
                                    "field_node_input_id": None,
                                    "operator": None,
                                    "value_node_input_id": None,
                                },
                            ],
                            "combinator": "AND",
                            "negated": False,
                            "field_node_input_id": None,
                            "operator": None,
                            "value_node_input_id": None,
                        },
                    },
                    {
                        "id": "c2fa8a44-923b-462a-b0d2-fa800a152e52",
                        "type": "ELSE",
                        "source_handle_id": "f9dde637-ea90-465f-a871-caf8380ae377",
                        "data": None,
                    },
                ],
                "version": "2",
            },
            "display_data": {"position": {"x": 200.0, "y": -50.0}},
            "base": {
                "name": "ConditionalNode",
                "module": ["vellum", "workflows", "nodes", "displayable", "conditional_node", "node"],
            },
            "definition": {
                "name": "CategoryConditionalNode",
                "module": ["tests", "workflows", "basic_conditional_node", "workflow"],
            },
            "trigger": {
                "id": "dd89e228-a23e-422b-80b2-34362c1c050e",
                "merge_behavior": "AWAIT_ANY",
            },
            "ports": [
                {
                    "id": "3a45b81f-95e4-4cbd-8997-bfdbe30251e8",
                    "name": "category_question",
                    "type": "IF",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {"type": "WORKFLOW_INPUT", "input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf"},
                        "operator": "=",
                        "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "question"}},
                    },
                },
                {
                    "id": "7202f702-1ebc-4067-ab1e-ec67e49158ee",
                    "name": "category_complaint",
                    "type": "ELIF",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {"type": "WORKFLOW_INPUT", "input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf"},
                        "operator": "=",
                        "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "complaint"}},
                    },
                },
                {
                    "id": "cf45705d-1a47-43a6-9d24-a7fdf78baae0",
                    "name": "category_compliment",
                    "type": "ELIF",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {"type": "WORKFLOW_INPUT", "input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf"},
                        "operator": "=",
                        "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "compliment"}},
                    },
                },
                {
                    "id": "f04610dd-61cf-41b0-b337-2235e101cdb0",
                    "name": "category_statement",
                    "type": "ELIF",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {
                            "type": "BINARY_EXPRESSION",
                            "lhs": {
                                "type": "WORKFLOW_INPUT",
                                "input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf",
                            },
                            "operator": "=",
                            "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "statement"}},
                        },
                        "operator": "and",
                        "rhs": {
                            "type": "BINARY_EXPRESSION",
                            "lhs": {
                                "type": "BINARY_EXPRESSION",
                                "lhs": {
                                    "type": "WORKFLOW_INPUT",
                                    "input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf",
                                },
                                "operator": "=",
                                "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "statement"}},
                            },
                            "operator": "and",
                            "rhs": {
                                "type": "BINARY_EXPRESSION",
                                "lhs": {
                                    "type": "WORKFLOW_INPUT",
                                    "input_variable_id": "eece050a-432e-4a2c-8c87-9480397e4cbf",
                                },
                                "operator": "=",
                                "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "statement"}},
                            },
                        },
                    },
                },
                {
                    "id": "f9dde637-ea90-465f-a871-caf8380ae377",
                    "name": "category_fallthrough",
                    "type": "ELSE",
                    "expression": None,
                },
            ],
        },
        conditional_node,
        ignore_order=True,
    )

    passthrough_nodes = [node for node in workflow_raw_data["nodes"] if node["type"] == "GENERIC"]
    assert len(passthrough_nodes) == 5

    assert not DeepDiff(
        [
            {
                "id": "9c22ee47-01da-4e4e-863d-b4a6874bed66",
                "type": "TERMINAL",
                "data": {
                    "label": "Final Output",
                    "name": "statement",
                    "target_handle_id": "f02a8971-e9a4-4716-bfb4-d08f5614b5d8",
                    "output_id": "cdbe2adf-9951-409a-b9a8-b8b349037f4f",
                    "output_type": "STRING",
                    "node_input_id": "2e742a40-cbee-4e19-9269-c62dc4a9204e",
                },
                "inputs": [
                    {
                        "id": "2e742a40-cbee-4e19-9269-c62dc4a9204e",
                        "key": "node_input",
                        "value": {
                            "rules": [
                                {
                                    "type": "NODE_OUTPUT",
                                    "data": {
                                        "node_id": "ed7caf01-9ae7-47a3-b15a-16697abaf486",
                                        "output_id": "74ea6af1-8934-4e3c-b68d-b93092b4be73",
                                    },
                                }
                            ],
                            "combinator": "OR",
                        },
                    }
                ],
                "display_data": {"position": {"x": 600.0, "y": -50.0}},
                "base": {
                    "name": "FinalOutputNode",
                    "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
                },
                "definition": None,
            },
            {
                "id": "47f0931c-41f6-4b84-bf39-0c486941f599",
                "type": "TERMINAL",
                "data": {
                    "label": "Final Output",
                    "name": "compliment",
                    "target_handle_id": "a4d57adc-58c1-40c6-810b-ee5fd923bfc5",
                    "output_id": "93f2cb75-6fa2-4e46-9488-c0bcd29153c0",
                    "output_type": "STRING",
                    "node_input_id": "9ba792e6-55e3-4a14-8768-e8ef6955c934",
                },
                "inputs": [
                    {
                        "id": "9ba792e6-55e3-4a14-8768-e8ef6955c934",
                        "key": "node_input",
                        "value": {
                            "rules": [
                                {
                                    "type": "NODE_OUTPUT",
                                    "data": {
                                        "node_id": "8df781b1-ff28-48a5-98a2-d7d796b932b0",
                                        "output_id": "61c357a1-41d8-4adf-bfe1-ce615c4d7d23",
                                    },
                                }
                            ],
                            "combinator": "OR",
                        },
                    }
                ],
                "display_data": {"position": {"x": 600.0, "y": -550.0}},
                "base": {
                    "name": "FinalOutputNode",
                    "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
                },
                "definition": None,
            },
            {
                "id": "e3d29229-f746-4125-819e-f847acbed307",
                "type": "TERMINAL",
                "data": {
                    "label": "Final Output",
                    "name": "complaint",
                    "target_handle_id": "c5dd9bf5-9e18-4dbc-8c20-2c0baf969ebe",
                    "output_id": "f936ae31-ba15-4864-8961-86231022a4d7",
                    "output_type": "STRING",
                    "node_input_id": "47f426a4-8770-4f30-a285-5d21849063a5",
                },
                "inputs": [
                    {
                        "id": "47f426a4-8770-4f30-a285-5d21849063a5",
                        "key": "node_input",
                        "value": {
                            "rules": [
                                {
                                    "type": "NODE_OUTPUT",
                                    "data": {
                                        "node_id": "68c02b7c-5077-4087-803d-841474a8081f",
                                        "output_id": "0ec68ffe-cbb7-4dbb-aaff-f6025bd62efa",
                                    },
                                }
                            ],
                            "combinator": "OR",
                        },
                    }
                ],
                "display_data": {"position": {"x": 600.0, "y": 200.0}},
                "base": {
                    "name": "FinalOutputNode",
                    "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
                },
                "definition": None,
            },
            {
                "id": "6efa7b45-0580-406d-85aa-439117ba8021",
                "type": "TERMINAL",
                "data": {
                    "label": "Final Output",
                    "name": "fallthrough",
                    "target_handle_id": "2283cd2c-b077-4b5d-a96f-aa2cd6023eda",
                    "output_id": "62ad462f-f819-4940-99ab-b3f145507f57",
                    "output_type": "STRING",
                    "node_input_id": "d2d3e7cc-f6d6-4ac1-ad19-e6f52a75b38f",
                },
                "inputs": [
                    {
                        "id": "d2d3e7cc-f6d6-4ac1-ad19-e6f52a75b38f",
                        "key": "node_input",
                        "value": {
                            "rules": [
                                {
                                    "type": "NODE_OUTPUT",
                                    "data": {
                                        "node_id": "148c61bd-e8b0-4d4b-8734-b043a72b90ed",
                                        "output_id": "fafa0bde-8508-43d5-a9c8-db5d49f307f6",
                                    },
                                }
                            ],
                            "combinator": "OR",
                        },
                    }
                ],
                "display_data": {"position": {"x": 600.0, "y": -300.0}},
                "base": {
                    "name": "FinalOutputNode",
                    "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
                },
                "definition": None,
            },
            {
                "id": "fa11b84b-1d76-4adc-ab28-cbbaa933c267",
                "type": "TERMINAL",
                "data": {
                    "label": "Final Output",
                    "name": "question",
                    "target_handle_id": "e1a6da28-02c5-40d7-8ac5-9fb07e2e3e1d",
                    "output_id": "c05f7d96-59a0-4d58-93d7-d451afd3f630",
                    "output_type": "STRING",
                    "node_input_id": "a13377c4-ae1c-4be4-b695-f9da590486ef",
                },
                "inputs": [
                    {
                        "id": "a13377c4-ae1c-4be4-b695-f9da590486ef",
                        "key": "node_input",
                        "value": {
                            "rules": [
                                {
                                    "type": "NODE_OUTPUT",
                                    "data": {
                                        "node_id": "0d959311-c836-4641-a867-58f63df9dfea",
                                        "output_id": "db9f7ff3-77e2-4b0a-9c39-bb4bb50e3ad5",
                                    },
                                }
                            ],
                            "combinator": "OR",
                        },
                    }
                ],
                "display_data": {"position": {"x": 600.0, "y": 450.0}},
                "base": {
                    "name": "FinalOutputNode",
                    "module": ["vellum", "workflows", "nodes", "displayable", "final_output_node", "node"],
                },
                "definition": None,
            },
        ],
        workflow_raw_data["nodes"][7:12],
        ignore_order=True,
    )

    # AND each edge should be serialized correctly
    serialized_edges = workflow_raw_data["edges"]
    assert not DeepDiff(
        [
            {
                "id": "32263c88-d725-4d03-a500-fadc34e10c9a",
                "source_node_id": "089b3201-537a-4ed7-8d15-2524a00e8534",
                "source_handle_id": "c2f0871d-0d9d-417f-8b0e-c813ccf880ac",
                "target_node_id": "9b619e4d-b0a7-4121-9060-100d457868cb",
                "target_handle_id": "dd89e228-a23e-422b-80b2-34362c1c050e",
                "type": "DEFAULT",
            },
            {
                "id": "1ae3cdb6-5b52-4ad3-bcfe-6997c86083f8",
                "source_node_id": "9b619e4d-b0a7-4121-9060-100d457868cb",
                "source_handle_id": "3a45b81f-95e4-4cbd-8997-bfdbe30251e8",
                "target_node_id": "0d959311-c836-4641-a867-58f63df9dfea",
                "target_handle_id": "7beba198-c452-4749-a38a-ea9420d84e14",
                "type": "DEFAULT",
            },
            {
                "id": "5704cb9c-9d01-4809-9d91-8014276e6574",
                "source_node_id": "9b619e4d-b0a7-4121-9060-100d457868cb",
                "source_handle_id": "7202f702-1ebc-4067-ab1e-ec67e49158ee",
                "target_node_id": "68c02b7c-5077-4087-803d-841474a8081f",
                "target_handle_id": "1dc4eebe-b6db-4229-96e5-115ff8cedb76",
                "type": "DEFAULT",
            },
            {
                "id": "c923e009-06c9-4978-b789-6ae995dcc81c",
                "source_node_id": "9b619e4d-b0a7-4121-9060-100d457868cb",
                "source_handle_id": "cf45705d-1a47-43a6-9d24-a7fdf78baae0",
                "target_node_id": "8df781b1-ff28-48a5-98a2-d7d796b932b0",
                "target_handle_id": "b73c39be-cbfe-4225-86e6-e6e4c161881e",
                "type": "DEFAULT",
            },
            {
                "id": "e487c031-fd5b-41b3-94d7-eb3f7ce8e25c",
                "source_node_id": "9b619e4d-b0a7-4121-9060-100d457868cb",
                "source_handle_id": "f04610dd-61cf-41b0-b337-2235e101cdb0",
                "target_node_id": "ed7caf01-9ae7-47a3-b15a-16697abaf486",
                "target_handle_id": "76fe7aec-5cd4-4c1a-b386-cfe09ebe66e4",
                "type": "DEFAULT",
            },
            {
                "id": "6a1c379d-bbe4-4034-8ac9-0353901ebc21",
                "source_node_id": "9b619e4d-b0a7-4121-9060-100d457868cb",
                "source_handle_id": "f9dde637-ea90-465f-a871-caf8380ae377",
                "target_node_id": "148c61bd-e8b0-4d4b-8734-b043a72b90ed",
                "target_handle_id": "c88839af-3a79-4310-abbd-e1553d981dce",
                "type": "DEFAULT",
            },
            {
                "id": "8a554637-e382-4a66-9b77-4eadce45a25a",
                "source_node_id": "ed7caf01-9ae7-47a3-b15a-16697abaf486",
                "source_handle_id": "cde43aef-f607-4b5d-87f6-9238dd4a3a2b",
                "target_node_id": "9c22ee47-01da-4e4e-863d-b4a6874bed66",
                "target_handle_id": "f02a8971-e9a4-4716-bfb4-d08f5614b5d8",
                "type": "DEFAULT",
            },
            {
                "id": "af083f7d-226c-4341-bb6f-756f00846b42",
                "source_node_id": "68c02b7c-5077-4087-803d-841474a8081f",
                "source_handle_id": "ef032cf7-c8df-4a98-827c-386dd8a5a346",
                "target_node_id": "e3d29229-f746-4125-819e-f847acbed307",
                "target_handle_id": "c5dd9bf5-9e18-4dbc-8c20-2c0baf969ebe",
                "type": "DEFAULT",
            },
            {
                "id": "47758209-70cb-4f12-b71f-dc28df0f6d0b",
                "source_node_id": "0d959311-c836-4641-a867-58f63df9dfea",
                "source_handle_id": "69a2121d-fc21-47a1-af49-6200aad836de",
                "target_node_id": "fa11b84b-1d76-4adc-ab28-cbbaa933c267",
                "target_handle_id": "e1a6da28-02c5-40d7-8ac5-9fb07e2e3e1d",
                "type": "DEFAULT",
            },
            {
                "id": "f08a49f8-8bfd-4c05-8f28-dfa536654af8",
                "source_node_id": "8df781b1-ff28-48a5-98a2-d7d796b932b0",
                "source_handle_id": "aeb6805d-2c9f-4d52-a690-341ea0e869b3",
                "target_node_id": "47f0931c-41f6-4b84-bf39-0c486941f599",
                "target_handle_id": "a4d57adc-58c1-40c6-810b-ee5fd923bfc5",
                "type": "DEFAULT",
            },
            {
                "id": "c45e03b4-dba6-4620-bc02-3847ad90086b",
                "source_node_id": "148c61bd-e8b0-4d4b-8734-b043a72b90ed",
                "source_handle_id": "26f50353-85ae-462f-b82d-9fd736900bd6",
                "target_node_id": "6efa7b45-0580-406d-85aa-439117ba8021",
                "target_handle_id": "2283cd2c-b077-4b5d-a96f-aa2cd6023eda",
                "type": "DEFAULT",
            },
        ],
        serialized_edges,
        ignore_order=True,
    )

    # AND the display data should be what we expect
    display_data = workflow_raw_data["display_data"]
    assert display_data == {
        "viewport": {
            "x": 0.0,
            "y": 0.0,
            "zoom": 1.0,
        }
    }

    # AND the definition should be what we expect
    definition = workflow_raw_data["definition"]
    assert definition == {
        "name": "CategoryWorkflow",
        "module": [
            "tests",
            "workflows",
            "basic_conditional_node",
            "workflow",
        ],
    }


def descriptors_with_lhs_and_rhs():
    return [
        (EqualsExpression(lhs="123", rhs="123"), "="),
        (DoesNotEqualExpression(lhs="123", rhs="123"), "!="),
        (LessThanExpression(lhs="123", rhs="123"), "<"),
        (GreaterThanExpression(lhs="123", rhs="123"), ">"),
        (LessThanOrEqualToExpression(lhs="123", rhs="123"), "<="),
        (GreaterThanOrEqualToExpression(lhs="123", rhs="123"), ">="),
        (ContainsExpression(lhs="123", rhs="123"), "contains"),
        (BeginsWithExpression(lhs="123", rhs="123"), "beginsWith"),
        (EndsWithExpression(lhs="123", rhs="123"), "endsWith"),
        (DoesNotContainExpression(lhs="123", rhs="123"), "doesNotContain"),
        (DoesNotBeginWithExpression(lhs="123", rhs="123"), "doesNotBeginWith"),
        (DoesNotEndWithExpression(lhs="123", rhs="123"), "doesNotEndWith"),
        (InExpression(lhs="123", rhs="123"), "in"),
        (NotInExpression(lhs="123", rhs="123"), "notIn"),
    ]


def descriptors_with_expression():
    return [
        (IsNullExpression(expression="123"), "null"),
        (IsNotNullExpression(expression="123"), "notNull"),
    ]


def descriptors_with_value_and_start_and_end():
    return [
        (BetweenExpression(value="123", start="123", end="123"), "between"),
        (NotBetweenExpression(value="123", start="123", end="123"), "notBetween"),
    ]


@pytest.mark.parametrize("descriptor, operator", descriptors_with_lhs_and_rhs())
def test_conditional_node_serialize_all_operators_with_lhs_and_rhs(descriptor, operator):
    # GIVEN a simple workflow with one conditional node
    workflow_cls = create_simple_workflow(descriptor)

    workflow_display = get_workflow_display(workflow_class=workflow_cls)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND the conditional node should be what we expect
    conditional_node = workflow_raw_data["nodes"][1]
    assert not DeepDiff(
        {
            "id": "a9143814-6bb0-4cb3-a817-4fc076417121",
            "type": "CONDITIONAL",
            "inputs": [
                {
                    "id": "738a274f-962d-466e-9aee-7774d3e05ab9",
                    "key": "f497b2bf-7d35-43af-b162-ced2d8abd46f.field",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "123"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "f30bceb4-39bf-433a-9229-b6871dbdbe00",
                    "key": "f497b2bf-7d35-43af-b162-ced2d8abd46f.value",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "123"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
            ],
            "data": {
                "label": "Simple Conditional Node",
                "target_handle_id": "c6e99e94-bc8e-47a4-b75c-cc96c6bedbb0",
                "conditions": [
                    {
                        "id": "a4c32611-fd58-4b98-9d08-313cfd1c214e",
                        "type": "IF",
                        "source_handle_id": "2ff87aa6-37cf-43dd-af9d-13b9198ab70a",
                        "data": {
                            "id": "650e7105-3e76-43ca-858f-b290970b438b",
                            "rules": [
                                {
                                    "id": "f497b2bf-7d35-43af-b162-ced2d8abd46f",
                                    "rules": None,
                                    "combinator": None,
                                    "negated": False,
                                    "field_node_input_id": "738a274f-962d-466e-9aee-7774d3e05ab9",
                                    "operator": f"{operator}",
                                    "value_node_input_id": "f30bceb4-39bf-433a-9229-b6871dbdbe00",
                                }
                            ],
                            "combinator": "AND",
                            "negated": False,
                            "field_node_input_id": None,
                            "operator": None,
                            "value_node_input_id": None,
                        },
                    },
                    {
                        "id": "342e5497-ea2b-4e5c-99cf-e6492f133a3c",
                        "type": "ELSE",
                        "source_handle_id": "9a7e8c2e-0228-4321-8f74-61cb5778f3df",
                        "data": None,
                    },
                ],
                "version": "2",
            },
            "display_data": {"position": {"x": 200.0, "y": -50.0}},
            "base": {
                "name": "ConditionalNode",
                "module": ["vellum", "workflows", "nodes", "displayable", "conditional_node", "node"],
            },
            "definition": {
                "name": "SimpleConditionalNode",
                "module": ["tests", "workflows", "basic_conditional_node", "workflow_with_only_one_conditional_node"],
            },
            "trigger": {
                "id": "c6e99e94-bc8e-47a4-b75c-cc96c6bedbb0",
                "merge_behavior": "AWAIT_ANY",
            },
            "ports": [
                {
                    "id": "2ff87aa6-37cf-43dd-af9d-13b9198ab70a",
                    "name": "text_str",
                    "type": "IF",
                    "expression": {
                        "type": "BINARY_EXPRESSION",
                        "lhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "123"}},
                        "operator": f"{operator}",
                        "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "123"}},
                    },
                },
                {
                    "id": "9a7e8c2e-0228-4321-8f74-61cb5778f3df",
                    "name": "text_fallthrough",
                    "type": "ELSE",
                    "expression": None,
                },
            ],
        },
        conditional_node,
        ignore_order=True,
    )


@pytest.mark.parametrize("descriptor, operator", descriptors_with_expression())
def test_conditional_node_serialize_all_operators_with_expression(descriptor, operator):
    # GIVEN a simple workflow with one conditional node
    workflow_cls = create_simple_workflow(descriptor)

    workflow_display = get_workflow_display(workflow_class=workflow_cls)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND the conditional node should be what we expect
    conditional_node = workflow_raw_data["nodes"][1]
    assert not DeepDiff(
        {
            "id": "a9143814-6bb0-4cb3-a817-4fc076417121",
            "type": "CONDITIONAL",
            "inputs": [
                {
                    "id": "738a274f-962d-466e-9aee-7774d3e05ab9",
                    "key": "f497b2bf-7d35-43af-b162-ced2d8abd46f.field",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "123"},
                            }
                        ],
                        "combinator": "OR",
                    },
                }
            ],
            "data": {
                "label": "Simple Conditional Node",
                "target_handle_id": "c6e99e94-bc8e-47a4-b75c-cc96c6bedbb0",
                "conditions": [
                    {
                        "id": "a4c32611-fd58-4b98-9d08-313cfd1c214e",
                        "type": "IF",
                        "source_handle_id": "2ff87aa6-37cf-43dd-af9d-13b9198ab70a",
                        "data": {
                            "id": "650e7105-3e76-43ca-858f-b290970b438b",
                            "rules": [
                                {
                                    "id": "f497b2bf-7d35-43af-b162-ced2d8abd46f",
                                    "rules": None,
                                    "combinator": None,
                                    "negated": False,
                                    "field_node_input_id": "738a274f-962d-466e-9aee-7774d3e05ab9",
                                    "operator": f"{operator}",
                                    "value_node_input_id": "f30bceb4-39bf-433a-9229-b6871dbdbe00",
                                }
                            ],
                            "combinator": "AND",
                            "negated": False,
                            "field_node_input_id": None,
                            "operator": None,
                            "value_node_input_id": None,
                        },
                    },
                    {
                        "id": "342e5497-ea2b-4e5c-99cf-e6492f133a3c",
                        "type": "ELSE",
                        "source_handle_id": "9a7e8c2e-0228-4321-8f74-61cb5778f3df",
                        "data": None,
                    },
                ],
                "version": "2",
            },
            "display_data": {"position": {"x": 200.0, "y": -50.0}},
            "base": {
                "name": "ConditionalNode",
                "module": ["vellum", "workflows", "nodes", "displayable", "conditional_node", "node"],
            },
            "definition": {
                "name": "SimpleConditionalNode",
                "module": ["tests", "workflows", "basic_conditional_node", "workflow_with_only_one_conditional_node"],
            },
            "trigger": {
                "id": "c6e99e94-bc8e-47a4-b75c-cc96c6bedbb0",
                "merge_behavior": "AWAIT_ANY",
            },
            "ports": [
                {
                    "id": "2ff87aa6-37cf-43dd-af9d-13b9198ab70a",
                    "name": "text_str",
                    "type": "IF",
                    "expression": {
                        "type": "UNARY_EXPRESSION",
                        "lhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "123"}},
                        "operator": f"{operator}",
                    },
                },
                {
                    "id": "9a7e8c2e-0228-4321-8f74-61cb5778f3df",
                    "name": "text_fallthrough",
                    "type": "ELSE",
                    "expression": None,
                },
            ],
        },
        conditional_node,
        ignore_order=True,
    )


@pytest.mark.parametrize("descriptor, operator", descriptors_with_value_and_start_and_end())
def test_conditional_node_serialize_all_operators_with_value_and_start_and_end(descriptor, operator):
    # GIVEN a simple workflow with one conditional node
    workflow_cls = create_simple_workflow(descriptor)

    workflow_display = get_workflow_display(workflow_class=workflow_cls)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]

    # AND the conditional node should be what we expect
    conditional_node = workflow_raw_data["nodes"][1]
    assert not DeepDiff(
        {
            "id": "a9143814-6bb0-4cb3-a817-4fc076417121",
            "type": "CONDITIONAL",
            "inputs": [
                {
                    "id": "738a274f-962d-466e-9aee-7774d3e05ab9",
                    "key": "f497b2bf-7d35-43af-b162-ced2d8abd46f.field",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "123"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
                {
                    "id": "f30bceb4-39bf-433a-9229-b6871dbdbe00",
                    "key": "f497b2bf-7d35-43af-b162-ced2d8abd46f.value",
                    "value": {
                        "rules": [
                            {
                                "type": "CONSTANT_VALUE",
                                "data": {"type": "STRING", "value": "123,123"},
                            }
                        ],
                        "combinator": "OR",
                    },
                },
            ],
            "data": {
                "label": "Simple Conditional Node",
                "target_handle_id": "c6e99e94-bc8e-47a4-b75c-cc96c6bedbb0",
                "conditions": [
                    {
                        "id": "a4c32611-fd58-4b98-9d08-313cfd1c214e",
                        "type": "IF",
                        "source_handle_id": "2ff87aa6-37cf-43dd-af9d-13b9198ab70a",
                        "data": {
                            "id": "650e7105-3e76-43ca-858f-b290970b438b",
                            "rules": [
                                {
                                    "id": "f497b2bf-7d35-43af-b162-ced2d8abd46f",
                                    "rules": None,
                                    "combinator": None,
                                    "negated": False,
                                    "field_node_input_id": "738a274f-962d-466e-9aee-7774d3e05ab9",
                                    "operator": f"{operator}",
                                    "value_node_input_id": "f30bceb4-39bf-433a-9229-b6871dbdbe00",
                                }
                            ],
                            "combinator": "AND",
                            "negated": False,
                            "field_node_input_id": None,
                            "operator": None,
                            "value_node_input_id": None,
                        },
                    },
                    {
                        "id": "342e5497-ea2b-4e5c-99cf-e6492f133a3c",
                        "type": "ELSE",
                        "source_handle_id": "9a7e8c2e-0228-4321-8f74-61cb5778f3df",
                        "data": None,
                    },
                ],
                "version": "2",
            },
            "display_data": {"position": {"x": 200.0, "y": -50.0}},
            "base": {
                "name": "ConditionalNode",
                "module": ["vellum", "workflows", "nodes", "displayable", "conditional_node", "node"],
            },
            "definition": {
                "name": "SimpleConditionalNode",
                "module": ["tests", "workflows", "basic_conditional_node", "workflow_with_only_one_conditional_node"],
            },
            "trigger": {
                "id": "c6e99e94-bc8e-47a4-b75c-cc96c6bedbb0",
                "merge_behavior": "AWAIT_ANY",
            },
            "ports": [
                {
                    "id": "2ff87aa6-37cf-43dd-af9d-13b9198ab70a",
                    "name": "text_str",
                    "type": "IF",
                    "expression": {
                        "type": "TERNARY_EXPRESSION",
                        "base": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "123"}},
                        "operator": f"{operator}",
                        "lhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "123"}},
                        "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "123"}},
                    },
                },
                {
                    "id": "9a7e8c2e-0228-4321-8f74-61cb5778f3df",
                    "name": "text_fallthrough",
                    "type": "ELSE",
                    "expression": None,
                },
            ],
        },
        conditional_node,
        ignore_order=True,
    )



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel, update_forward_refs
from .test_suite_run_exec_config import TestSuiteRunExecConfig
from .test_suite_run_progress import TestSuiteRunProgress
from .test_suite_run_state import TestSuiteRunState
from .test_suite_run_test_suite import TestSuiteRunTestSuite


class TestSuiteRunRead(UniversalBaseModel):
    id: str
    created: dt.datetime
    test_suite: TestSuiteRunTestSuite
    state: TestSuiteRunState = pydantic.Field()
    """
    The current state of this run
    
    * `QUEUED` - Queued
    * `RUNNING` - Running
    * `COMPLETE` - Complete
    * `FAILED` - Failed
    * `CANCELLED` - Cancelled
    """

    exec_config: typing.Optional[TestSuiteRunExecConfig] = pydantic.Field(default=None)
    """
    Configuration that defines how the Test Suite should be run
    """

    progress: typing.Optional[TestSuiteRunProgress] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


from .array_vellum_value import ArrayVellumValue  # noqa: E402, F401, I001

update_forward_refs(TestSuiteRunRead)



import pytest
import json

from vellum import SearchResponse, SearchResult, SearchResultDocument
from vellum.client.types.chat_message import ChatMessage
from vellum.client.types.json_vellum_value_request import JsonVellumValueRequest
from vellum.client.types.search_filters_request import SearchFiltersRequest
from vellum.client.types.search_request_options_request import SearchRequestOptionsRequest
from vellum.client.types.search_result_merging_request import SearchResultMergingRequest
from vellum.client.types.search_weights_request import SearchWeightsRequest
from vellum.client.types.string_vellum_value_request import StringVellumValueRequest
from vellum.client.types.vellum_value_logical_condition_group_request import VellumValueLogicalConditionGroupRequest
from vellum.client.types.vellum_value_logical_condition_request import VellumValueLogicalConditionRequest
from vellum.workflows.errors import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes.displayable.bases.types import (
    MetadataLogicalCondition,
    MetadataLogicalConditionGroup,
    SearchFilters,
)
from vellum.workflows.nodes.displayable.search_node.node import SearchNode
from vellum.workflows.state.base import BaseState


def test_run_workflow__happy_path(vellum_client):
    """Confirm that we can successfully invoke a Workflow with the new option attributes"""

    # GIVEN a workflow that's set up run a Search Node
    class MySearchNode(SearchNode):
        query = "Search query"
        document_index = "document_index"
        limit = 1
        weights = SearchWeightsRequest(
            semantic_similarity=0.8,
            keywords=0.2,
        )
        result_merging = SearchResultMergingRequest(
            enabled=True,
        )
        filters = SearchFilters(
            external_ids=["external_id"],
            metadata=MetadataLogicalConditionGroup(
                combinator="AND",
                negated=False,
                conditions=[
                    MetadataLogicalCondition(
                        lhs_variable="TYPE",
                        operator="=",
                        rhs_variable="COMPANY",
                    ),
                    MetadataLogicalCondition(
                        lhs_variable="NAME",
                        operator="notNull",
                    ),
                ],
            ),
        )

    # AND a Search request that will return a 200 ok resposne
    search_response = SearchResponse(
        results=[
            SearchResult(
                text="Search query", score="0.0", keywords=["keywords"], document=SearchResultDocument(label="label")
            )
        ]
    )

    vellum_client.search.return_value = search_response

    # WHEN we run the workflow
    outputs = MySearchNode().run()

    # THEN the workflow should have completed successfully
    assert outputs.text == "Search query"

    # AND the options should be as expected
    assert vellum_client.search.call_args.kwargs["options"] == SearchRequestOptionsRequest(
        limit=1,
        weights=SearchWeightsRequest(
            semantic_similarity=0.8,
            keywords=0.2,
        ),
        result_merging=SearchResultMergingRequest(
            enabled=True,
        ),
        filters=SearchFiltersRequest(
            external_ids=["external_id"],
            metadata=VellumValueLogicalConditionGroupRequest(
                combinator="AND",
                negated=False,
                conditions=[
                    VellumValueLogicalConditionRequest(
                        lhs_variable=StringVellumValueRequest(value="TYPE"),
                        operator="=",
                        rhs_variable=StringVellumValueRequest(value="COMPANY"),
                    ),
                    VellumValueLogicalConditionRequest(
                        lhs_variable=StringVellumValueRequest(value="NAME"),
                        operator="notNull",
                        rhs_variable=JsonVellumValueRequest(value=None),
                    ),
                ],
            ),
        ),
    )


def test_run_workflow__happy_path__options_attribute(vellum_client):
    """Confirm that we can successfully invoke a single Search node with the legacy options attribute"""

    # GIVEN a workflow that's set up run a Search Node
    class MySearchNode(SearchNode):
        query = "Search query"
        document_index = "document_index"
        options = SearchRequestOptionsRequest(
            limit=1,
            weights=SearchWeightsRequest(
                semantic_similarity=0.8,
                keywords=0.2,
            ),
            result_merging=SearchResultMergingRequest(
                enabled=True,
            ),
            filters=SearchFiltersRequest(
                external_ids=["external_id"],
                metadata=VellumValueLogicalConditionGroupRequest(
                    combinator="AND",
                    negated=False,
                    conditions=[
                        VellumValueLogicalConditionRequest(
                            lhs_variable=StringVellumValueRequest(value="TYPE"),
                            operator="=",
                            rhs_variable=StringVellumValueRequest(value="COMPANY"),
                        )
                    ],
                ),
            ),
        )

    # AND a Search request that will return a 200 ok resposne
    search_response = SearchResponse(
        results=[
            SearchResult(
                text="Search query", score="0.0", keywords=["keywords"], document=SearchResultDocument(label="label")
            )
        ]
    )

    vellum_client.search.return_value = search_response

    # WHEN we run the workflow
    outputs = MySearchNode().run()

    # THEN the workflow should have completed successfully
    assert outputs.text == "Search query"

    # AND the options should be as expected
    assert vellum_client.search.call_args.kwargs["options"] == SearchRequestOptionsRequest(
        limit=1,
        weights=SearchWeightsRequest(
            semantic_similarity=0.8,
            keywords=0.2,
        ),
        result_merging=SearchResultMergingRequest(
            enabled=True,
        ),
        filters=SearchFiltersRequest(
            external_ids=["external_id"],
            metadata=VellumValueLogicalConditionGroupRequest(
                combinator="AND",
                negated=False,
                conditions=[
                    VellumValueLogicalConditionRequest(
                        lhs_variable=StringVellumValueRequest(value="TYPE"),
                        operator="=",
                        rhs_variable=StringVellumValueRequest(value="COMPANY"),
                    )
                ],
            ),
        ),
    )


def test_run_workflow__chat_history_as_query(vellum_client):
    """
    Confirm that we can successfully invoke a Search node with a chat history as the query param,
    backwards compatible with original workflows
    """

    # GIVEN a state definition with a chat history
    class MyState(BaseState):
        chat_history: list[ChatMessage]

    # AND a Search Node that uses the chat history as the query param
    class MySearchNode(SearchNode[MyState]):
        query = MyState.chat_history  # type: ignore[assignment]
        document_index = "document_index"
        limit = 1

    # AND a Search request that will return a 200 ok resposne
    search_response = SearchResponse(
        results=[
            SearchResult(
                text="Search query", score="0.0", keywords=["keywords"], document=SearchResultDocument(label="label")
            )
        ]
    )

    vellum_client.search.return_value = search_response

    # WHEN we run the workflow
    outputs = MySearchNode(state=MyState(chat_history=[ChatMessage(role="USER", text="Hello, world!")])).run()

    # THEN the workflow should have completed successfully
    assert outputs.text == "Search query"

    # AND the options should be as expected
    assert json.loads(vellum_client.search.call_args.kwargs["query"]) == [
        {"role": "USER", "text": "Hello, world!", "source": None, "content": None}
    ]


@pytest.mark.parametrize("invalid_query", [None, ""])
def test_run_workflow__invalid_query_raises_validation_error(invalid_query):
    """Confirm that missing/None/empty query raises proper user-facing validation error"""

    class MySearchNode(SearchNode):
        query = invalid_query  # type: ignore[assignment]
        document_index = "document_index"

    with pytest.raises(NodeException) as exc_info:
        MySearchNode().run()

    assert exc_info.value.code == WorkflowErrorCode.INVALID_INPUTS
    assert "query" in exc_info.value.message.lower()
    assert "required" in exc_info.value.message.lower() or "missing" in exc_info.value.message.lower()



# This file was auto-generated by Fern from our API Definition.

import typing

from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.request_options import RequestOptions
from ...types.workflow_execution_detail import WorkflowExecutionDetail
from .raw_client import AsyncRawWorkflowExecutionsClient, RawWorkflowExecutionsClient


class WorkflowExecutionsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawWorkflowExecutionsClient(client_wrapper=client_wrapper)

    @property
    def _client_wrapper(self) -> SyncClientWrapper:
        return self._raw_client._client_wrapper

    @property
    def with_raw_response(self) -> RawWorkflowExecutionsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawWorkflowExecutionsClient
        """
        return self._raw_client

    def retrieve_workflow_execution_detail(
        self, execution_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WorkflowExecutionDetail:
        """
        Parameters
        ----------
        execution_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowExecutionDetail


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.workflow_executions.retrieve_workflow_execution_detail(
            execution_id="execution_id",
        )
        """
        _response = self._raw_client.retrieve_workflow_execution_detail(execution_id, request_options=request_options)
        return _response.data


class AsyncWorkflowExecutionsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawWorkflowExecutionsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawWorkflowExecutionsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawWorkflowExecutionsClient
        """
        return self._raw_client

    async def retrieve_workflow_execution_detail(
        self, execution_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> WorkflowExecutionDetail:
        """
        Parameters
        ----------
        execution_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        WorkflowExecutionDetail


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.workflow_executions.retrieve_workflow_execution_detail(
                execution_id="execution_id",
            )


        asyncio.run(main())
        """
        _response = await self._raw_client.retrieve_workflow_execution_detail(
            execution_id, request_options=request_options
        )
        return _response.data



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .bad_request_error import BadRequestError
from .forbidden_error import ForbiddenError
from .internal_server_error import InternalServerError
from .not_found_error import NotFoundError
from .too_many_requests_error import TooManyRequestsError
from .unauthorized_error import UnauthorizedError

__all__ = [
    "BadRequestError",
    "ForbiddenError",
    "InternalServerError",
    "NotFoundError",
    "TooManyRequestsError",
    "UnauthorizedError",
]



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel, update_forward_refs
from .test_suite_run_external_exec_config_data import TestSuiteRunExternalExecConfigData


class TestSuiteRunExternalExecConfig(UniversalBaseModel):
    """
    Execution configuration for running a Vellum Test Suite against an external callable
    """

    data: TestSuiteRunExternalExecConfigData
    test_case_ids: typing.Optional[typing.List[str]] = pydantic.Field(default=None)
    """
    Optionally specify a subset of test case ids to run. If not provided, all test cases within the test suite will be run by default.
    """

    type: typing.Literal["EXTERNAL"] = "EXTERNAL"

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


from .array_vellum_value import ArrayVellumValue  # noqa: E402, F401, I001

update_forward_refs(TestSuiteRunExternalExecConfig)



# This file was auto-generated by Fern from our API Definition.

import typing

VellumErrorCodeEnum = typing.Union[
    typing.Literal[
        "INVALID_REQUEST",
        "INVALID_INPUTS",
        "PROVIDER_ERROR",
        "REQUEST_TIMEOUT",
        "INTERNAL_SERVER_ERROR",
        "USER_DEFINED_ERROR",
    ],
    typing.Any,
]



# This file was auto-generated by Fern from our API Definition.

import typing

from .chat_history_input_request import ChatHistoryInputRequest
from .json_input_request import JsonInputRequest
from .string_input_request import StringInputRequest

PromptDeploymentInputRequest = typing.Union[StringInputRequest, JsonInputRequest, ChatHistoryInputRequest]



# This file was auto-generated by Fern from our API Definition.

import typing

from .api_node_result import ApiNodeResult
from .code_execution_node_result import CodeExecutionNodeResult
from .conditional_node_result import ConditionalNodeResult
from .map_node_result import MapNodeResult
from .merge_node_result import MergeNodeResult
from .metric_node_result import MetricNodeResult
from .prompt_node_result import PromptNodeResult
from .search_node_result import SearchNodeResult
from .subworkflow_node_result import SubworkflowNodeResult
from .templating_node_result import TemplatingNodeResult
from .terminal_node_result import TerminalNodeResult

WorkflowNodeResultData = typing.Union[
    PromptNodeResult,
    SearchNodeResult,
    TemplatingNodeResult,
    CodeExecutionNodeResult,
    ConditionalNodeResult,
    ApiNodeResult,
    TerminalNodeResult,
    MergeNodeResult,
    SubworkflowNodeResult,
    MetricNodeResult,
    MapNodeResult,
]



import re
from typing import Optional, TypeVar, Union, overload

from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.references.node import NodeReference

_T = TypeVar("_T")


@overload
def raise_if_descriptor(node_attr: BaseDescriptor[_T]) -> _T: ...


@overload
def raise_if_descriptor(node_attr: _T) -> _T: ...


def raise_if_descriptor(node_attr: Union[NodeReference[_T], _T]) -> Optional[_T]:
    if not isinstance(node_attr, NodeReference):
        raise AttributeError(f"Expected to find a Node descriptor, but found '{node_attr.__class__.__name__}'")

    return node_attr.instance


def to_kebab_case(string: str) -> str:
    """Converts a string from Capital Case to kebab-case."""
    string = re.sub(r"(\s|_|-)+", " ", string)
    string = re.sub(
        r"[A-Z]{2,}(?=[A-Z][a-z]+[0-9]*|\b)|[A-Z]?[a-z]+[0-9]*|[A-Z]|[0-9]+",
        lambda mo: " " + mo.group(0).lower(),
        string,
    )
    return "-".join(string.split())



# This file was auto-generated by Fern from our API Definition.

import typing

from .execution_array_vellum_value import ExecutionArrayVellumValue
from .execution_chat_history_vellum_value import ExecutionChatHistoryVellumValue
from .execution_error_vellum_value import ExecutionErrorVellumValue
from .execution_function_call_vellum_value import ExecutionFunctionCallVellumValue
from .execution_json_vellum_value import ExecutionJsonVellumValue
from .execution_number_vellum_value import ExecutionNumberVellumValue
from .execution_search_results_vellum_value import ExecutionSearchResultsVellumValue
from .execution_string_vellum_value import ExecutionStringVellumValue
from .execution_thinking_vellum_value import ExecutionThinkingVellumValue

ExecutionVellumValue = typing.Union[
    ExecutionStringVellumValue,
    ExecutionNumberVellumValue,
    ExecutionJsonVellumValue,
    ExecutionChatHistoryVellumValue,
    ExecutionSearchResultsVellumValue,
    ExecutionErrorVellumValue,
    ExecutionArrayVellumValue,
    ExecutionFunctionCallVellumValue,
    ExecutionThinkingVellumValue,
]



from dataclasses import dataclass, field
from uuid import UUID
from typing import Optional, Type

from pydantic import Field

from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum.workflows.workflows.base import BaseWorkflow
from vellum_ee.workflows.display.editor.types import NodeDisplayData


class WorkflowDisplayDataViewport(UniversalBaseModel):
    x: float = 0.0
    y: float = 0.0
    zoom: float = 1.0


class WorkflowDisplayData(UniversalBaseModel):
    viewport: WorkflowDisplayDataViewport = Field(default_factory=WorkflowDisplayDataViewport)


@dataclass
class WorkflowMetaDisplay:
    entrypoint_node_id: UUID
    entrypoint_node_source_handle_id: UUID
    entrypoint_node_display: NodeDisplayData = Field(default_factory=NodeDisplayData)
    display_data: WorkflowDisplayData = field(default_factory=WorkflowDisplayData)

    @classmethod
    def get_default(cls, workflow: Type[BaseWorkflow]) -> "WorkflowMetaDisplay":
        entrypoint_node_id = uuid4_from_hash(f"{workflow.__id__}|entrypoint_node_id")
        entrypoint_node_source_handle_id = uuid4_from_hash(f"{workflow.__id__}|entrypoint_node_source_handle_id")

        return WorkflowMetaDisplay(
            entrypoint_node_id=entrypoint_node_id,
            entrypoint_node_source_handle_id=entrypoint_node_source_handle_id,
            entrypoint_node_display=NodeDisplayData(),
        )


@dataclass
class WorkflowInputsDisplay:
    id: UUID
    name: Optional[str] = None
    color: Optional[str] = None


@dataclass
class StateValueDisplay:
    id: UUID
    name: Optional[str] = None
    color: Optional[str] = None


@dataclass
class EdgeDisplay:
    id: UUID


@dataclass
class EntrypointDisplay:
    id: UUID
    edge_display: EdgeDisplay


@dataclass
class WorkflowOutputDisplay:
    id: UUID
    name: str



# This file was auto-generated by Fern from our API Definition.

import typing

from .test_suite_test_case_created_bulk_result import TestSuiteTestCaseCreatedBulkResult
from .test_suite_test_case_deleted_bulk_result import TestSuiteTestCaseDeletedBulkResult
from .test_suite_test_case_rejected_bulk_result import TestSuiteTestCaseRejectedBulkResult
from .test_suite_test_case_replaced_bulk_result import TestSuiteTestCaseReplacedBulkResult

TestSuiteTestCaseBulkResult = typing.Union[
    TestSuiteTestCaseCreatedBulkResult,
    TestSuiteTestCaseReplacedBulkResult,
    TestSuiteTestCaseDeletedBulkResult,
    TestSuiteTestCaseRejectedBulkResult,
]



# This file was auto-generated by Fern from our API Definition.

# isort: skip_file

from .document_indexes_list_request_status import DocumentIndexesListRequestStatus

__all__ = ["DocumentIndexesListRequestStatus"]
